@@ -5,9 +5,8 @@ django-linguo by Zach Mathew
 
 https://github.com/zmathew/django-linguo
 """
-from __future__ import with_statement  # Python 2.5 compatibility
 from django.db import models
-from django.db.models.fields.related import RelatedField
+from django.db.models.fields.related import RelatedField, RelatedObject
 from django.db.models.sql.where import Constraint
 from django.utils.tree import Node
 
@@ -19,37 +18,32 @@ from modeltranslation.utils import (build_localized_fieldname, get_language,
 def get_translatable_fields_for_model(model):
     from modeltranslation.translator import NotRegistered, translator
     try:
-        return translator.get_options_for_model(model).fields
+        return translator.get_options_for_model(model).get_field_names()
     except NotRegistered:
         return None
 
 
 def rewrite_lookup_key(model, lookup_key):
+    pieces = lookup_key.split('__', 1)
+    original_key = pieces[0]
+
     translatable_fields = get_translatable_fields_for_model(model)
     if translatable_fields is not None:
-        pieces = lookup_key.split('__')
         # If we are doing a lookup on a translatable field,
         # we want to rewrite it to the actual field name
         # For example, we want to rewrite "name__startswith" to "name_fr__startswith"
         if pieces[0] in translatable_fields:
-            lookup_key = build_localized_fieldname(pieces[0], get_language())
-            remaining_lookup = '__'.join(pieces[1:])
-            if remaining_lookup:
-                lookup_key = '%s__%s' % (lookup_key, remaining_lookup)
+            pieces[0] = build_localized_fieldname(pieces[0], get_language())
 
-    pieces = lookup_key.split('__')
     if len(pieces) > 1:
         # Check if we are doing a lookup to a related trans model
         fields_to_trans_models = get_fields_to_translatable_models(model)
         for field_to_trans, transmodel in fields_to_trans_models:
-            if pieces[0] == field_to_trans:
-                sub_lookup = '__'.join(pieces[1:])
-                if sub_lookup:
-                    sub_lookup = rewrite_lookup_key(transmodel, sub_lookup)
-                    lookup_key = '%s__%s' % (pieces[0], sub_lookup)
+            # Check ``original key``, as pieces[0] may have been already rewritten.
+            if original_key == field_to_trans:
+                pieces[1] = rewrite_lookup_key(transmodel, pieces[1])
                 break
-
-    return lookup_key
+    return '__'.join(pieces)
 
 
 def rewrite_order_lookup_key(model, lookup_key):
@@ -58,15 +52,24 @@ def rewrite_order_lookup_key(model, lookup_key):
     else:
         return rewrite_lookup_key(model, lookup_key)
 
+_F2TM_CACHE = {}
+
 
 def get_fields_to_translatable_models(model):
-    results = []
-    for field_name in model._meta.get_all_field_names():
-        field_object, modelclass, direct, m2m = model._meta.get_field_by_name(field_name)
-        if direct and isinstance(field_object, RelatedField):
-            if get_translatable_fields_for_model(field_object.related.parent_model) is not None:
-                results.append((field_name, field_object.related.parent_model))
-    return results
+    if model not in _F2TM_CACHE:
+        results = []
+        for field_name in model._meta.get_all_field_names():
+            field_object, modelclass, direct, m2m = model._meta.get_field_by_name(field_name)
+            # Direct relationship
+            if direct and isinstance(field_object, RelatedField):
+                if get_translatable_fields_for_model(field_object.related.parent_model) is not None:
+                    results.append((field_name, field_object.related.parent_model))
+            # Reverse relationship
+            if isinstance(field_object, RelatedObject):
+                if get_translatable_fields_for_model(field_object.model) is not None:
+                    results.append((field_name, field_object.model))
+        _F2TM_CACHE[model] = results
+    return _F2TM_CACHE[model]
 
 
 class MultilingualQuerySet(models.query.QuerySet):
@@ -87,10 +90,15 @@ class MultilingualQuerySet(models.query.QuerySet):
                 self.query.add_ordering(*ordering)
 
     # This method was not present in django-linguo
-    def _clone(self, *args, **kwargs):
+    def _clone(self, klass=None, *args, **kwargs):
+        if klass is not None and not issubclass(klass, MultilingualQuerySet):
+            class NewClass(klass, MultilingualQuerySet):
+                pass
+            NewClass.__name__ = 'Multilingual%s' % klass.__name__
+            klass = NewClass
         kwargs.setdefault('_rewrite', self._rewrite)
         kwargs.setdefault('_populate', self._populate)
-        return super(MultilingualQuerySet, self)._clone(*args, **kwargs)
+        return super(MultilingualQuerySet, self)._clone(klass, *args, **kwargs)
 
     # This method was not present in django-linguo
     def rewrite(self, mode=True):
@@ -123,7 +131,8 @@ class MultilingualQuerySet(models.query.QuerySet):
                 c.field = self.model._meta.get_field(new_name)
                 c.col = c.field.column
         if isinstance(q, Node):
-            map(self._rewrite_where, q.children)
+            for child in q.children:
+                self._rewrite_where(child)
 
     def _rewrite_order(self):
         self.query.order_by = [rewrite_order_lookup_key(self.model, field_name)
@@ -135,7 +144,7 @@ class MultilingualQuerySet(models.query.QuerySet):
         if isinstance(q, tuple) and len(q) == 2:
             return rewrite_lookup_key(self.model, q[0]), q[1]
         if isinstance(q, Node):
-            q.children = map(self._rewrite_q, q.children)
+            q.children = list(map(self._rewrite_q, q.children))
         return q
 
     # This method was not present in django-linguo
@@ -147,7 +156,7 @@ class MultilingualQuerySet(models.query.QuerySet):
             q.name = rewrite_lookup_key(self.model, q.name)
             return q
         if isinstance(q, Node):
-            q.children = map(self._rewrite_f, q.children)
+            q.children = list(map(self._rewrite_f, q.children))
         return q
 
     def _filter_or_exclude(self, negate, *args, **kwargs):
@@ -206,6 +215,57 @@ class MultilingualQuerySet(models.query.QuerySet):
         with auto_populate(self._populate_mode):
             return super(MultilingualQuerySet, self).get_or_create(**kwargs)
 
+    def _append_translated(self, fields):
+        "If translated field is encountered, add also all its translation fields."
+        fields = set(fields)
+        from modeltranslation.translator import translator
+        opts = translator.get_options_for_model(self.model)
+        for key, translated in opts.fields.items():
+            if key in fields:
+                fields = fields.union(f.name for f in translated)
+        return fields
+
+    # This method was not present in django-linguo
+    def defer(self, *fields):
+        fields = self._append_translated(fields)
+        return super(MultilingualQuerySet, self).defer(*fields)
+
+    # This method was not present in django-linguo
+    def only(self, *fields):
+        fields = self._append_translated(fields)
+        return super(MultilingualQuerySet, self).only(*fields)
+
+    # This method was not present in django-linguo
+    def raw_values(self, *fields):
+        return super(MultilingualQuerySet, self).values(*fields)
+
+    # This method was not present in django-linguo
+    def values(self, *fields):
+        if not self._rewrite:
+            return super(MultilingualQuerySet, self).values(*fields)
+        new_args = []
+        for key in fields:
+            new_args.append(rewrite_lookup_key(self.model, key))
+        vqs = super(MultilingualQuerySet, self).values(*new_args)
+        vqs.field_names = list(fields)
+        return vqs
+
+    # This method was not present in django-linguo
+    def values_list(self, *fields, **kwargs):
+        if not self._rewrite:
+            return super(MultilingualQuerySet, self).values_list(*fields, **kwargs)
+        new_args = []
+        for key in fields:
+            new_args.append(rewrite_lookup_key(self.model, key))
+        return super(MultilingualQuerySet, self).values_list(*new_args, **kwargs)
+
+    # This method was not present in django-linguo
+    def dates(self, field_name, *args, **kwargs):
+        if not self._rewrite:
+            return super(MultilingualQuerySet, self).dates(field_name, *args, **kwargs)
+        new_key = rewrite_lookup_key(self.model, field_name)
+        return super(MultilingualQuerySet, self).dates(new_key, *args, **kwargs)
+
 
 class MultilingualManager(models.Manager):
     use_for_related_fields = True
@@ -216,6 +276,9 @@ class MultilingualManager(models.Manager):
     def populate(self, *args, **kwargs):
         return self.get_query_set().populate(*args, **kwargs)
 
+    def raw_values(self, *args, **kwargs):
+        return self.get_query_set().raw_values(*args, **kwargs)
+
     def get_query_set(self):
         qs = super(MultilingualManager, self).get_query_set()
         if qs.__class__ == models.query.QuerySet:
