@@ -1,10 +1,15 @@
 # -*- coding: utf-8 -*-
 from django.conf import settings
-from django.db.models import Manager
+from django.utils.six import with_metaclass
+from django.db.models import Manager, ForeignKey, OneToOneField
 from django.db.models.base import ModelBase
+from django.db.models.signals import post_init
+from django.dispatch import receiver
 
 from modeltranslation import settings as mt_settings
-from modeltranslation.fields import TranslationFieldDescriptor, create_translation_field
+from modeltranslation.fields import (NONE, create_translation_field, TranslationFieldDescriptor,
+                                     TranslatedRelationIdDescriptor,
+                                     LanguageCacheSingleObjectDescriptor)
 from modeltranslation.manager import MultilingualManager, rewrite_lookup_key
 from modeltranslation.utils import build_localized_fieldname
 
@@ -34,7 +39,7 @@ class FieldsAggregationMetaClass(type):
         return super(FieldsAggregationMetaClass, cls).__new__(cls, name, bases, attrs)
 
 
-class TranslationOptions(object):
+class TranslationOptions(with_metaclass(FieldsAggregationMetaClass, object)):
     """
     Translatable fields are declared by registering a model using
     ``TranslationOptions`` class with appropriate ``fields`` attribute.
@@ -49,8 +54,11 @@ class TranslationOptions(object):
     in between in the inheritance chain), while ``fields`` also includes fields
     inherited from concrete supermodels (giving all translated fields available
     on a model).
+
+    ``related`` attribute inform whether this model is related part of some relation
+    with translated model. This model may be not translated itself.
+    ``related_fields`` contains names of reverse lookup fields.
     """
-    __metaclass__ = FieldsAggregationMetaClass
 
     def __init__(self, model):
         """
@@ -58,14 +66,16 @@ class TranslationOptions(object):
         """
         self.model = model
         self.registered = False
+        self.related = False
         self.local_fields = dict((f, set()) for f in self.fields)
         self.fields = dict((f, set()) for f in self.fields)
+        self.related_fields = []
 
     def update(self, other):
         """
         Update with options from a superclass.
         """
-        if other.model._meta.abstract:
+        if other.model._meta.abstract or self.model._meta.proxy:
             self.local_fields.update(other.local_fields)
         self.fields.update(other.fields)
 
@@ -76,6 +86,12 @@ class TranslationOptions(object):
         self.local_fields[field].add(translation_field)
         self.fields[field].add(translation_field)
 
+    def get_field_names(self):
+        """
+        Return name of all fields that can be used in filtering.
+        """
+        return list(self.fields.keys()) + self.related_fields
+
     def __str__(self):
         local = tuple(self.local_fields.keys())
         inherited = tuple(set(self.fields.keys()) - set(local))
@@ -89,7 +105,7 @@ def add_translation_fields(model, opts):
 
     Adds newly created translation fields to the given translation options.
     """
-    for field_name in opts.local_fields.iterkeys():
+    for field_name in opts.local_fields.keys():
         for l in settings.LANGUAGES:
             # Create a dynamic translation field
             translation_field = create_translation_field(
@@ -110,21 +126,20 @@ def add_translation_fields(model, opts):
 def add_manager(model):
     """
     Monkey patches the original model to use MultilingualManager instead of
-    default manager (``objects``).
+    default managers (not only ``objects``, but also every manager defined and inherited).
 
-    If model has a custom manager, then merge it with MultilingualManager.
+    Custom managers are merged with MultilingualManager.
     """
-    if not hasattr(model, 'objects'):
-        return
-    current_manager = model.objects
-    if isinstance(current_manager, MultilingualManager):
-        return
-    if current_manager.__class__ is Manager:
-        current_manager.__class__ = MultilingualManager
-    else:
-        class NewMultilingualManager(MultilingualManager, current_manager.__class__):
-            pass
-        current_manager.__class__ = NewMultilingualManager
+    for _, attname, cls in model._meta.concrete_managers:
+        current_manager = getattr(model, attname)
+        if isinstance(current_manager, MultilingualManager):
+            continue
+        if current_manager.__class__ is Manager:
+            current_manager.__class__ = MultilingualManager
+        else:
+            class NewMultilingualManager(MultilingualManager, current_manager.__class__):
+                pass
+            current_manager.__class__ = NewMultilingualManager
 
 
 def patch_constructor(model):
@@ -134,29 +149,80 @@ def patch_constructor(model):
     old_init = model.__init__
 
     def new_init(self, *args, **kwargs):
-        populate_translation_fields(self.__class__, kwargs)
-        for key, val in kwargs.items():
-            new_key = rewrite_lookup_key(model, key)
-            # Old key is intentionally left in case old_init wants to play with it
-            kwargs.setdefault(new_key, val)
+        self._mt_init = True
+        if not self._deferred:
+            populate_translation_fields(self.__class__, kwargs)
+            for key, val in list(kwargs.items()):
+                new_key = rewrite_lookup_key(model, key)
+                # Old key is intentionally left in case old_init wants to play with it
+                kwargs.setdefault(new_key, val)
         old_init(self, *args, **kwargs)
     model.__init__ = new_init
 
 
+@receiver(post_init)
+def delete_mt_init(sender, instance, **kwargs):
+    if hasattr(instance, '_mt_init'):
+        del instance._mt_init
+
+
+def patch_clean_fields(model):
+    """
+    Patch clean_fields method to handle different form types submission.
+    """
+    old_clean_fields = model.clean_fields
+
+    def new_clean_fields(self, exclude=None):
+        if hasattr(self, '_mt_form_pending_clear'):
+            # Some form translation fields has been marked as clearing value.
+            # Check if corresponding translated field was also saved (not excluded):
+            # - if yes, it seems like form for MT-unaware app. Ignore clearing (left value from
+            #   translated field unchanged), as if field was omitted from form
+            # - if no, then proceed as normally: clear the field
+            for field_name, value in self._mt_form_pending_clear.items():
+                orig_field_name = self._meta.get_field(field_name).translated_field.name
+                if orig_field_name in exclude:
+                    setattr(self, field_name, value)
+            delattr(self, '_mt_form_pending_clear')
+        old_clean_fields(self, exclude)
+    model.clean_fields = new_clean_fields
+
+
+def patch_metaclass(model):
+    """
+    Monkey patches original model metaclass to exclude translated fields on deferred subclasses.
+    """
+    old_mcs = model.__class__
+
+    class translation_deferred_mcs(old_mcs):
+        """
+        This metaclass is essential for deferred subclasses (obtained via only/defer) to work.
+
+        When deferred subclass is created, some translated fields descriptors could be overridden
+        by DeferredAttribute - which would cause translation retrieval to fail.
+        Prevent this from happening with deleting those attributes from class being created.
+        This metaclass would be called from django.db.models.query_utils.deferred_class_factory
+        """
+        def __new__(cls, name, bases, attrs):
+            if attrs.get('_deferred', False):
+                opts = translator.get_options_for_model(model)
+                for field_name in opts.fields.keys():
+                    attrs.pop(field_name, None)
+            return super(translation_deferred_mcs, cls).__new__(cls, name, bases, attrs)
+    # Assign to __metaclass__ wouldn't work, since metaclass search algorithm check for __class__.
+    # http://docs.python.org/2/reference/datamodel.html#__metaclass__
+    model.__class__ = translation_deferred_mcs
+
+
 def delete_cache_fields(model):
     opts = model._meta
-    try:
-        del opts._field_cache
-    except AttributeError:
-        pass
-    try:
-        del opts._field_name_cache
-    except AttributeError:
-        pass
-    try:
-        del opts._name_map
-    except AttributeError:
-        pass
+    cached_attrs = ('_field_cache', '_field_name_cache', '_name_map', 'fields', 'concrete_fields',
+                    'local_concrete_fields')
+    for attr in cached_attrs:
+        try:
+            delattr(opts, attr)
+        except AttributeError:
+            pass
 
 
 def populate_translation_fields(sender, kwargs):
@@ -192,7 +258,7 @@ def populate_translation_fields(sender, kwargs):
         populate = 'all'
 
     opts = translator.get_options_for_model(sender)
-    for key, val in kwargs.items():
+    for key, val in list(kwargs.items()):
         if key in opts.fields:
             if populate == 'all':
                 # Set the value for every language.
@@ -209,6 +275,17 @@ def populate_translation_fields(sender, kwargs):
                 raise AttributeError("Unknown population mode '%s'." % populate)
 
 
+def patch_related_object_descriptor_caching(ro_descriptor):
+    """
+    Patch SingleRelatedObjectDescriptor or ReverseSingleRelatedObjectDescriptor to use
+    language-aware caching.
+    """
+    class NewSingleObjectDescriptor(LanguageCacheSingleObjectDescriptor, ro_descriptor.__class__):
+        pass
+    ro_descriptor.accessor = ro_descriptor.related.get_accessor_name()
+    ro_descriptor.__class__ = NewSingleObjectDescriptor
+
+
 class Translator(object):
     """
     A Translator object encapsulates an instance of a translator. Models are
@@ -267,21 +344,50 @@ class Translator(object):
             # Patch __init__ to rewrite fields
             patch_constructor(model)
 
+            # Patch clean_fields to verify form field clearing
+            patch_clean_fields(model)
+
+            # Patch __metaclass__ to allow deferring to work
+            patch_metaclass(model)
+
             # Substitute original field with descriptor
-            model_fallback_values = getattr(opts, 'fallback_values', None)
             model_fallback_languages = getattr(opts, 'fallback_languages', None)
-            for field_name in opts.local_fields.iterkeys():
-                if model_fallback_values is None:
-                    field_fallback_value = None
-                elif isinstance(model_fallback_values, dict):
-                    field_fallback_value = model_fallback_values.get(field_name, None)
+            model_fallback_values = getattr(opts, 'fallback_values', NONE)
+            model_fallback_undefined = getattr(opts, 'fallback_undefined', NONE)
+            for field_name in opts.local_fields.keys():
+                field = model._meta.get_field(field_name)
+                if isinstance(model_fallback_values, dict):
+                    field_fallback_value = model_fallback_values.get(field_name, NONE)
                 else:
                     field_fallback_value = model_fallback_values
+                if isinstance(model_fallback_undefined, dict):
+                    field_fallback_undefined = model_fallback_undefined.get(field_name, NONE)
+                else:
+                    field_fallback_undefined = model_fallback_undefined
                 descriptor = TranslationFieldDescriptor(
-                    model._meta.get_field(field_name),
+                    field,
+                    fallback_languages=model_fallback_languages,
                     fallback_value=field_fallback_value,
-                    fallback_languages=model_fallback_languages)
+                    fallback_undefined=field_fallback_undefined)
                 setattr(model, field_name, descriptor)
+                if isinstance(field, ForeignKey):
+                    # We need to use a special descriptor so that
+                    # _id fields on translated ForeignKeys work
+                    # as expected.
+                    desc = TranslatedRelationIdDescriptor(field_name, model_fallback_languages)
+                    setattr(model, field.get_attname(), desc)
+
+                    # Set related field names on other model
+                    if not field.rel.is_hidden():
+                        other_opts = self._get_options_for_model(field.rel.to)
+                        other_opts.related = True
+                        other_opts.related_fields.append(field.related_query_name())
+                        add_manager(field.rel.to)  # Add manager in case of non-registered model
+
+                if isinstance(field, OneToOneField):
+                    # Fix translated_field caching for SingleRelatedObjectDescriptor
+                    sro_descriptor = getattr(field.rel.to, field.related.get_accessor_name())
+                    patch_related_object_descriptor_caching(sro_descriptor)
 
     def unregister(self, model_or_iterable):
         """
@@ -298,7 +404,7 @@ class Translator(object):
             self.get_options_for_model(model)
             # Invalidate all submodels options and forget about
             # the model itself.
-            for desc, desc_opts in self._registry.items():
+            for desc, desc_opts in list(self._registry.items()):
                 if not issubclass(desc, model):
                     continue
                 if model != desc and desc_opts.registered:
@@ -349,7 +455,7 @@ class Translator(object):
         semantic of throwing exception for models not directly registered.
         """
         opts = self._get_options_for_model(model)
-        if not opts.registered:
+        if not opts.registered and not opts.related:
             raise NotRegistered('The model "%s" is not registered for '
                                 'translation' % model.__name__)
         return opts
