@@ -1,7 +1,8 @@
 # -*- coding: utf-8 -*-
 from contextlib import contextmanager
 
-from django.utils.encoding import force_unicode
+from django.utils import six
+from django.utils.encoding import force_text
 from django.utils.translation import get_language as _get_language
 from django.utils.functional import lazy
 
@@ -33,8 +34,8 @@ def build_localized_fieldname(field_name, lang):
 
 
 def _build_localized_verbose_name(verbose_name, lang):
-    return u'%s [%s]' % (force_unicode(verbose_name), lang)
-build_localized_verbose_name = lazy(_build_localized_verbose_name, unicode)
+    return force_text('%s [%s]') % (force_text(verbose_name), lang)
+build_localized_verbose_name = lazy(_build_localized_verbose_name, six.text_type)
 
 
 def _join_css_class(bits, offset):
